import os
import sys
from concurrent.futures.thread import ThreadPoolExecutor
import threading

import requests

from multissh.connection import Connection
from multissh.utils import parse_credentials, print_error, print_str, get_input


class MultiSSH:
    __running = True
    __verbose = False
    __source = None
    __workers = None
    __connections = []
    __thread_pool = []

    def __init__(self, config):
        self.__verbose = config['verbose']
        self.__source = config['source']
        self.__workers = config['workers']

    def run(self):
        credentials = []

        if self.__source.startswith('http://') or self.__source.startswith('https://'):
            try:
                response = requests.get(self.__source, headers={'User-agent': 'MultiSSH'})

                for line in response.text().split('\n'):
                    data = line.strip()

                    if data != "":
                        credentials.append(parse_credentials(data))

            except requests.exceptions.ConnectionError as e:
                print_error('Unable to download credentials: %s' % str(e))
            except Exception as e:
                print_error('Something went wrong: %s' % str(e))
        else:
            if os.path.exists(self.__source):
                with open(self.__source, 'r') as f:
                    for line in f.readlines():
                        data = line.strip()

                        if data != "":
                            credentials.append(parse_credentials(data))
            else:
                print_error('Credentials file: %s does not exist!' % self.__source)
                sys.exit(0)

        if len(credentials) == 0:
            print_error('There are no credentials')
            sys.exit(0)

        if not self.__workers:
            self.__workers = len(credentials)

        self.connect_to_hosts(credentials)

        timer = threading.Timer(60, self.health_check)
        timer.start()

        while self.__running:
            try:
                command = get_input()
            except KeyboardInterrupt:
                command = 'exit'

            if command == 'exit':
                self.__running = False
                self.disconnect_from_hosts()
            else:
                self.execute_cmd(command)

        print_str('\nBye bye!')
        sys.exit(0)

    def connect_to_hosts(self, credentials):
        for data in credentials:
            self.__connections.append(Connection(data['host'], data['port'], data['username'], data['password']))

        with ThreadPoolExecutor(max_workers=self.__workers) as executor:
            for connection in self.__connections:
                executor.submit(connection.connect)

    def disconnect_from_hosts(self):
        with ThreadPoolExecutor(max_workers=self.__workers) as executor:
            for connection in self.__connections:
                executor.submit(connection.disconnect)

    def execute_cmd(self, command):
        with ThreadPoolExecutor(max_workers=self.__workers) as executor:
            for connection in self.__connections:
                executor.submit(connection.exec_cmd, command)

    def health_check(self):
        with ThreadPoolExecutor(max_workers=self.__workers) as executor:
            for connection in self.__connections:
                executor.submit(connection.check_connection)




import paramiko
from paramiko import SSHClient, AuthenticationException, SSHException
from multissh.utils import print_success, print_error, get_hostname, print_str, parse_response
import time


class Connection:
    __hostname = None
    __host = None
    __port = None
    __username = None
    __password = None

    __connection = None

    def __init__(self, host=None, port=None, username=None, password=None):
        self.__host = host
        self.__port = port
        self.__username = username
        self.__password = password

    def connect(self, max_tries=5, timeout=3):
        tries = 0

        while tries < max_tries:
            try:
                self.__connection = SSHClient()
                self.__connection.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                self.__connection.connect(hostname=self.__host, port=int(self.__port), username=self.__username, password=self.__password)

                if self.__connection.get_transport().is_active():
                    stdin, stdout, stderr = self.exec('hostname')

                    self.__hostname = get_hostname(('\n'.join(stdout)).strip())

                    print_success('Successfully connected!', self.__hostname)
                    return True
                else:
                    print_error('Connection failed!', self.__host)
                    return False

            except AuthenticationException:
                print_error('Authentication error!', self.__host)
                return False

            except SSHException as e:
                print_error(str(e), self.__host)
                return False

            except Exception as e:
                print_error('Connection failed with exception: %s' % str(e), self.__host)
                time.sleep(timeout * 1000)
                tries += 1

        return False

    def disconnect(self):
        self.exec('exit')
        self.__connection.close()

    def exec_cmd(self, command):
        stdin, stdout, stderr = self.exec(command)

        stderr = parse_response(stderr)
        stdout = parse_response(stdout)

        if stderr != "":
            print_error(stderr, self.__hostname)

        if stdout:
            print_str(stdout, 'info', self.__hostname)

    def exec(self, command, buffer_size=-1, timeout=10, get_pty=False, environment={}):
        if self.check_connection():
            return self.__connection.exec_command(command, bufsize=buffer_size, timeout=timeout, get_pty=get_pty, environment=environment)
        else:
            print_error('Unable to send command - connection closed!', self.__hostname)

        return False

    def check_connection(self):
        if not self.__connection.get_transport().is_active():
            print_error('Connection is closed, reconnecting...', self.__hostname)
            return self.connect()

        return True



import setuptools

with open('README.md', 'r') as f:
    long_description = f.read()

setuptools.setup(
    name='multissh',
    version='0.1.4',
    author='Jaan Porkon',
    author_email='jaantrill@gmail.com',
    description='Tool to connect to multiple SSH servers and broadcast commands.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/JaanPorkon/MultiSSH',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    install_requires=['requests', 'paramiko'],
    scripts=['bin/mssh']
)


from tldextract import tldextract


def get_hostname(hostname):
    h_data = tldextract.extract(hostname)

    if h_data.subdomain:
        hostname = h_data.subdomain
    else:
        if h_data.registered_domain:
            hostname = h_data.registered_domain

    return hostname


def print_str(message, str_type='info', hostname=None):
    if str_type == 'info':
        color = ConsoleColor.WHITE
    elif str_type == 'success':
        color = ConsoleColor.GREEN
    elif str_type == 'error':
        color = ConsoleColor.RED
    else:
        color = ConsoleColor.WHITE

    if hostname:
        print('%s[%s]: %s%s%s' % (ConsoleColor.HEADER, hostname, color, message, ConsoleColor.END))
    else:
        print('%s%s%s' % (color, message, ConsoleColor.END))


def print_error(message, hostname=None):
    print_str(message, 'error', hostname)


def print_success(message, hostname=None):
    print_str(message, 'success', hostname)


def parse_credentials(data):
    data = data.split(',')

    if len(data) != 4:
        print_error('Invalid credential line: %s' % ','.join(data))
        return False

    return {
        'host': data[0],
        'port': data[1],
        'username': data[2],
        'password': data[3]
    }


def parse_response(response):
    lines = []

    for line in response:
        line = line.strip()

        if line != '':
            lines.append(line)

    return '\n'.join(lines)


def get_input():
    return input('%s%s%s' % (ConsoleColor.GREEN, 'multissh$: ', ConsoleColor.END)).strip()


class ConsoleColor:
    GREEN = '\033[92m'
    RED = '\033[91m'
    HEADER = '\033[95m'
    END = '\033[0m'
    WHITE = '\33[37m'
